//  ComputationalAssembler.java
import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;
import swarm.collections.ListImpl;
import swarm.collections.ListIndex;
import swarm.random.NormalDistImpl;
import java.lang.Float;
import java.io.*;
import java.util.*;
/**
 * The ComputationalAssembler class instances make
 * computational processes; we have a computationa assembler instance for
 * each unit
 *
 *
 * @author<br/>
 * Marco Lamieri<a href="mailto:lamieri@econ.unito.it"></a></br>
 * Francesco Merlo<a href="mailto:merlo@econ.unito.it"></a></br>
 * @version 0.9.7.31.b (BasicJVE 1.0)
 */
public class ComputationalAssembler extends ComputationalAssemblerBasic
{

	//Our MemoryMatrixes
	MemoryMatrix designMatrix, basicNetMatrix, tcMatrix, orMatrix, agentMatrix;
	//A normal distribution for random numbers
	NormalDistImpl normal;
	//grossSales File management
	public boolean grossSalesLogFileOpen = false;
	FileWriter grossSalesLog=null;
	int samplesN=0;
	float sampleSales = 0;
	float itemSales = 0;
	public MyWriter report=null;
	/**
	* the constructor for ComputationalAssembler
	*/
	public ComputationalAssembler (Zone aZone, OrderDistiller od, ESFrameModelSwarm mo)
	{
		// Call the constructor for the parent class.
		super(aZone, od, mo);
		normal = new NormalDistImpl (getZone());
	}
	//***********************************
	// BasicJES Computational codes
	//***********************************

	/** computational operations with code -1901 <br><br>
	*
	*  this computational code increment 1 by 1 position 0,0 of the
	*  unique received matrix and set the status to done
	*/
	public void c1901()
	{
		checkMatrixes(1901,1);

		designMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
		layer = pendingComputationalSpecificationSet.getOrderLayer();

		// Zeroing the counter if empty
		if(designMatrix.getEmpty(layer,0,0))
			designMatrix.setValue(layer,0,0,0.0);

		int counter = (int) designMatrix.getValue(layer,0,0);
		counter++;

		//Here we store the number of MetaSample produced
		designMatrix.setValue(layer,0,0,counter);

		System.out.println("MetaSample #" + counter + " with layer #" + layer + " has been drawn");

		done=true;
	}
	// end c1901

	/** computational operations with code -1902 <br><br>
	*
	*  this computational code dial with licensee's predictions.
	*
	*/
	public void c1902()
	{
		checkMatrixes(1902,2);

		designMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
		agentMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(1);
		layer=pendingComputationalSpecificationSet.getOrderLayer();

		// Making prediction for available MetaSample
		if(!designMatrix.getEmpty(layer,0,0))
		{
			//Switching to an other article
			int j = getCounter(agentMatrix,1);

			//The licensee prediction
			int forecast = (int) normal.getSampleWithMean$withVariance(500,900000);

			//Min value
			if (forecast < 10)
				forecast = 10;
			//Max value
			if (forecast > 25000)
				forecast = 25000;

			//Storing the prediction
			agentMatrix.setValue(layer,1,j,forecast);

			System.out.println("Licensee #" + myUnit.getUnitNumber() + " forecasted " + forecast + 
				" for MetaSample #" + j + " with layer #" + layer);
			
			// End of predictions
			done=true;
		}
		//Waiting for some MetaSample -> done=false
	}
	// end c1902

	/** computational operations with code -1903 <br><br>
	*
	*  this computational code sums up predictions.
	*
	*/
	public void c1903()
	{
		float sum = 0;
		int threshold = 18000;
		boolean endForecast = true;

		checkMatrixes(1901,37);

		designMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
		basicNetMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(36);
		layer = pendingComputationalSpecificationSet.getOrderLayer();

		// Getting the number of articles
		int counter = (int) designMatrix.getValue(layer,0,0);

		for (int j=1 ; j <=counter ; j++)
		{
			for (int t=1 ; t<=35 ; t++)
			{
				MemoryMatrix agentMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(t);
				//Checking if the licensee made a prediction for this MetaSample
				if(! (agentMatrix.getEmpty(layer,1,j)))
				{
					sum = sum + agentMatrix.getValue(layer,1,j);
				}
			}
			//Storing total predictions in basicNet Matrix
			basicNetMatrix.setValue(layer,1,j,sum);
			//Checking if MetaSample can be produced
			if (sum > threshold)
			{
				basicNetMatrix.setValue(layer,2,j,1);
				System.out.println("MetaSample #" + j + " with layer #" + layer +
					" can be produced, forecast=" + sum + " status=1");
			}
			else
			{
				basicNetMatrix.setValue(layer,2,j,0);
				System.out.println("MetaSample #" + j + " with layer #" + layer +
					" can NOT be produced, forecast=" + sum + " status=0");
			}
			//Waiting all predictions
			for (int t=1 ; t<=35 ; t++)
			{
				MemoryMatrix agentMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(t);

				if(agentMatrix.getEmpty(layer,1,j))
					endForecast = false;
			}

			if (endForecast)
			{
				System.out.println("MetaSample #" + j + " with layer #" + layer + " forecasted by all licensee");
				done=true;
			}
			sum = 0;
		}
	}
	// end c1903

	/** computational operations with code -1904 <br><br>
	*
	*  this computational code check treshold and if above
	*  set 1 to row 2 matrix in pos 1 *  and set done=true.
	*  1 cell for each number of row read in matrix in pos 0 (0,0),
	*  whit a counter in position (0,1)
	*/
	public void c1904()
	{
		float val;

		checkMatrixes(1904,2);

		designMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
		basicNetMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(1);
		layer = pendingComputationalSpecificationSet.getOrderLayer();

		//Switching to an other article
		int i = getCounter(basicNetMatrix, 1);

		//check the number of items and the status
		if(!designMatrix.getEmpty(layer,0,0) && !basicNetMatrix.getEmpty(layer,2,i))
		{
			if(i <= designMatrix.getValue(layer,0,0) && basicNetMatrix.getValue(layer,2,i) == 1)
			{
				System.out.println("Unit #" + myUnit.getUnitNumber() + " produced MetaSample #" + i 
					+ " with layer #" + layer + ", status=2");
				
				basicNetMatrix.setValue(layer,2,i,2);
				
				done=true;
			}
			else
			{
				System.out.println("Unit #" + myUnit.getUnitNumber() + " NOT produced MetaSample #" + i 
					+ " with layer #" + layer + ", status=1" );
			}
		}
	}
	// end c1904

	/** computational operations with code -1905 <br><br>
	*
	* This computational code dials with TC offer-price
	*/
	public void c1905()
	{
		float offerPrice, random;

		checkMatrixes(1905,3);

		designMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
		basicNetMatrix	= (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(1);
		tcMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(2);
		layer = pendingComputationalSpecificationSet.getOrderLayer();

		//Making an offer-price for available articles
		if(!designMatrix.getEmpty(layer,0,0))
		{
			int counter = (int) designMatrix.getValue(layer,0,0);

			for (int i=1; i<=counter; i++)
			{
				if(!basicNetMatrix.getEmpty(layer,2,i))
				{
					if(basicNetMatrix.getValue(layer,2,i) > 0)
					{
						//The offer-price
						offerPrice= (int) normal.getSampleWithMean$withVariance(7,10);
						//Min value
						if (offerPrice<1)
							offerPrice = 1;
						//Max value
						if (offerPrice>20)
							offerPrice=20;
						//Storing the offer-price of the TC
   					tcMatrix.setValue(layer, 1,i, offerPrice);

						System.out.println("Unit #" + myUnit.getUnitNumber() + " offer price for MetaSample #" + i 
							+  " with layer #" + layer + " is $" + offerPrice);
   				}
   				else
   					System.out.println("Unit #" + myUnit.getUnitNumber() + " can not make an offer for Metasample #" + i
							+  " with layer #" + layer );
				}
			}
			done=true;
		}
		// Waiting for some MetaSample -> done=false
	}
	// end c1905


	/** computational operations with code -1906 <br><br>
	* This computational code dial with the bidding phase
	*/
	public void c1906()
	{
		MemoryMatrix choiceMatrix;

		checkMatrixes(1906,8);

		designMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
		basicNetMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(6);
		orMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(7);
		layer = pendingComputationalSpecificationSet.getOrderLayer();

		// Checking MetaSamples
		if(!designMatrix.getEmpty(layer,0,0))
		{
			int counter = (int) designMatrix.getValue(layer,0,0);

			for (int i=1; i<=counter; i++)
			{
				int choice = 0;
				float offer=0;
				//Checking the status
				if(!basicNetMatrix.getEmpty(layer,2,i))
				{
					if(basicNetMatrix.getValue(layer,2,i) > 0)
					{
						//Here we make the bidding
						for (int t=1 ; t<=5 ; t++)
						{
							MemoryMatrix tcMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(t);

							if(!tcMatrix.getEmpty(layer,1,i))
							{
								if(tcMatrix.getValue(layer,1,i) < offer || offer==0)
								{
									offer = tcMatrix.getValue(layer,1,i);
									choice = t;
								}
							}
						}
						//Checking the TC chosen
						if(!basicNetMatrix.getEmpty(layer,2,i))
						{
							if(basicNetMatrix.getValue(layer,2,i)>0 && (!(choice==0)))
							{
								//Getting its offer price ...
								choiceMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(choice);

								float offerPrice = choiceMatrix.getValue(layer,1,i);
								//... storing it in basicNet Matrix ...
								basicNetMatrix.setValue(layer,3,i,offerPrice);
								//... and in the orMatrix
								orMatrix.setValue(layer,1,i,choice);

								System.out.println("Trading Company #" + orMatrix.getValue(layer,1,i) + 
									" has been chosen for MetaSample #" + i + " with layer #" + layer + " price " 
										+ basicNetMatrix.getValue(layer,3,i));
							}
						}
					}
				}
			}
			done=true;
		}
		// Waiting for some MetaSample -> done=false
	}
	// end c1906

	/** computational operations with code -1907 <br><br>
	*
	*  this computational code dial with licensee's first order.
	*
	*/
	public void c1907()
	{
		checkMatrixes(1907,3);

		designMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
		agentMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(1);
		basicNetMatrix=(MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(2);
		layer = pendingComputationalSpecificationSet.getOrderLayer();

		if(!designMatrix.getEmpty(layer,0,0))
		{
			int j = getCounter(agentMatrix,2);
			if (!basicNetMatrix.getEmpty(layer,2,j))
			{
				//Check if the licensee received the related MetaSample
				if (basicNetMatrix.getValue(layer,2,j)>1)
				{
					int order1= (int) normal.getSampleWithMean$withVariance(500,900000);
					//Min value
					if (order1<10)
						order1 = 10;
					//Max value
					if (order1>25000)
						order1 = 25000;

					agentMatrix.setValue(layer,2,j,order1);

					if (basicNetMatrix.getEmpty(layer,4,j))
						basicNetMatrix.setValue(layer,4,j,0.0);

					int totOrder1 = (int) basicNetMatrix.getValue(layer,4,j) + order1;
					basicNetMatrix.setValue(layer,4,j,totOrder1);

					System.out.println("The 1st order of Licensee #" + myUnit.getUnitNumber() + " is " + order1 + " for item #" + 
						j + " with layer #" + layer + " totOrder #"+totOrder1);
					
					done=true;
				}
				// The article can not be ordered -> done=false
			}
		}
	}
	// end c1907

	/** computational operations with code -1908 <br><br>
	*
	*  this computational code dial with or branch choice.
	*
	*/
	public void c1908()
	{
		checkMatrixes(1908,3);

		designMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
		basicNetMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(1);
		orMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(2);
		layer = pendingComputationalSpecificationSet.getOrderLayer();

		if(!designMatrix.getEmpty(layer,0,0))
		{
			//Switching to another article
			int i = getCounter(orMatrix, 0);

			if(i <= designMatrix.getValue(layer,0,0))
			{
				if(!basicNetMatrix.getEmpty(layer,2,i))
				{
					//Check if the item can be produced (status>1)
				    if(basicNetMatrix.getValue(layer,2,i) > 1 && !orMatrix.getEmpty(layer,1,i))
					{
						float branch = orMatrix.getValue(layer,1,i);
						orMatrix.setValue(layer,1,0,branch);

						System.out.println("Branch  #" + orMatrix.getValue(layer,1,0) +
							" produced item #" + i + " with layer #" + layer);

						basicNetMatrix.setValue(layer,2,i,3.0);

						done= true;
					}
				}
				// The article can not be produced -> done=false
			}
		}
	}
	//end c1908

	/** computational operations with code -1909 <br><br>
	*
	*  this computational code dial with licensee's second order.
	*
	*/
	public void c1909()
	{
	    int threshold = 100;

		checkMatrixes(1909,3);

		designMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
                agentMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(1);
		basicNetMatrix=(MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(2);
		layer = pendingComputationalSpecificationSet.getOrderLayer();

                int j = getCounter(agentMatrix,3);
	        
                if (!designMatrix.getEmpty(layer,0,0))
	        {
                  if (j <= designMatrix.getValue(layer,0,0))
		  {
                      //Checks if the licensee has already done the first order
   		      if (!agentMatrix.getEmpty(layer,2,j))
		      {
		          int select = (int) normal.getSampleWithMean$withVariance(100,50000);
                       
                          //decide to do or not another order
                          if (select > threshold)
                          {         
		             int order2 = (int) normal.getSampleWithMean$withVariance(500,900000);
			    
                              //Min value
			      if (order2<10)
			          order2 = 10;
			   
                              //Max value
			      if (order2>25000)
			          order2 = 25000;

	    		      agentMatrix.setValue(layer,3,j,order2);

  			      if (basicNetMatrix.getEmpty(layer,5,j))
			          basicNetMatrix.setValue(layer,5,j,0.0);

                              int totOrder2 = (int)  basicNetMatrix.getValue(layer,5,j) + order2;
			  
                              basicNetMatrix.setValue(layer,5,j,totOrder2);

	  	              System.out.println("The 2nd order of Licensee #" + myUnit.getUnitNumber() + " is " + order2 + " for item #" + 
		 	    	                  j + " with layer #" + layer + " totOrder #"+ totOrder2);
					
			  }//end if(select)
		        
                          else
                          {
                              agentMatrix.setValue(layer,3,j,0.0);
                            
                              if (basicNetMatrix.getEmpty(layer,5,j))
			          basicNetMatrix.setValue(layer,5,j,0.0);			   

                          }//end else
			                                
			
                              int totOrder1 = (int) basicNetMatrix.getValue(layer,4,j);
                              int totOrder2 = (int) basicNetMatrix.getValue(layer,5,j);
			      int totOrder = totOrder1 + totOrder2;
             
			      basicNetMatrix.setValue(layer,6,j,totOrder);
                   
		       }//end if (checking first order)
		   }
		}
			      done = true;
                       // The article can not be ordered -> done=false
                   
	}// end c1909

	/** computational operations with code -1910 <br><br>
	*
	*  This computational code prints the Report and stops the simulation.
	*
	*/
	public void c1910()
       	{
		checkMatrixes(1910,2);

		designMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
		basicNetMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(1);
		layer = pendingComputationalSpecificationSet.getOrderLayer();

		try
		{
			report = new MyWriter("log/report.txt");
			report.write("Layer | MetaSampleN | Status0 | Status1 | Status2 | Status3 | Status 4  \n");
		}
		catch (IOException e)
		{
			System.out.println(e);
			MyExit.exit(1);
		}

		for(int ln=1; ln<=16; ln++)
		{
			int metaSamplesN=0;
			int[] status;
			status = new int[5];

			for(int s=0; s<=4; s++)
				status[s]=0;

			if(!designMatrix.getEmpty(ln, 0, 0))
				metaSamplesN= (int) designMatrix.getValue(ln, 0, 0);

			for(int item=1; item<=150; item++)
			{
				if(!basicNetMatrix.getEmpty(ln, 2, item))
					status[(int) basicNetMatrix.getValue(ln, 2,  item)]++;
			}
			String text = ln + " | " + metaSamplesN + " | " + status[0] + " | " + status[1] + " | " + status[2] 
				+ " | " + status[3] + " | " + status[4] + " \n";

			try
			{
				report.write(text);
			}
			catch(IOException e)
			{
				System.out.println(e);
				MyExit.exit(1);
			}
		}
		try
		{
			report.close();
		}
		catch(IOException e)
		{
			System.out.println(e);
			MyExit.exit(1);
		}

		System.out.println("***** THIS IS THE END! *****");
		StartESFrame.eSFrameObserverSwarm.getControlPanel().setStateStopped();

		done = true;
		

        }//end c1910
       
	/** computational operations with code -1913 <br><br>
	*
	*  This computational code distributes items to the licensees.
	*
	*/
	public void c1913()
	{
	   	checkMatrixes(1913,2);

		designMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
		basicNetMatrix = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(1);
	        layer = pendingComputationalSpecificationSet.getOrderLayer();

                int counter = getCounter(basicNetMatrix,2);

                if (!designMatrix.getEmpty(layer,0,0))
		{
                   if(counter <= designMatrix.getValue(layer,0,0))
		   {             
                      //Check if item has been produced	
                      if(basicNetMatrix.getValue(layer,2,counter) > 2)
                      {
                         System.out.println("Item  #" + counter + " with layer #" + layer + " has been delivered");
                                    
                         basicNetMatrix.setValue(layer,2,counter,4.0);
      	              } 
		   }
		} 
		                        
                      done = true;
            	    
	}//end c1913
	
//************************
// Private functions
//************************

	//The counter is used in some computationl code in order to take actions
	// on different article with only a recipe
	private int getCounter(MemoryMatrix currentMatrix, int row)
	{
		//Zeroing the counter if empty
		if(currentMatrix.getEmpty(layer,row,0))
			currentMatrix.setValue(layer,row,0,0.0);
		//Here we increment the counter
		int counter = (int) currentMatrix.getValue(layer,row,0);
		counter++;
		//The counter must be less than the number of MetaSample
		if(counter <= designMatrix.getValue(layer,0,0))
			currentMatrix.setValue(layer,row,0,counter);

		return counter;
	}

	private void checkMatrixes(int code, int numberOfMatrixes)
	{
		if(pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed()!=numberOfMatrixes)
		{
			System.out.println("Code -"+ code + " requires " + numberOfMatrixes + " matrix; " 
				+ pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed() + " found in order # " +
				pendingComputationalSpecificationSet.getOrderNumber());

			MyExit.exit(1);
		}
	}

	private void logOpen(boolean eraseFile, String fileName)
	{
		File fileOut = new File(fileName);
		if(!this.grossSalesLogFileOpen && eraseFile)
		{
			try
			{
				FileWriter erase  = new FileWriter(fileOut, false);
				erase.close();
			}
			catch (IOException e)
			{
				System.out.println(e);
				MyExit.exit(1);
			}
		}
		try
		{
			grossSalesLog = new FileWriter(fileOut, true);
			this.grossSalesLogFileOpen = true ;
		}
		catch (IOException e)
		{
			System.out.println(e);
			MyExit.exit(1);
		}
	}
}
